/*
 * This file is part of Baritone.
 *
 * Baritone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Baritone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Baritone.  If not, see <https://www.gnu.org/licenses/>.
 */

package baritone.behavior;

import baritone.Baritone;
import baritone.api.Settings;
import baritone.api.behavior.ILookBehavior;
import baritone.api.event.events.PlayerUpdateEvent;
import baritone.api.event.events.RotationMoveEvent;
import baritone.api.utils.Rotation;
import net.minecraft.util.math.MathHelper;

public final class LookBehavior extends Behavior implements ILookBehavior {

    /**
     * Target's values are as follows:
     * <p>
     * getFirst() -> yaw
     * getSecond() -> pitch
     */
    private Rotation target;

    /**
     * Whether or not rotations are currently being forced
     */
    private boolean force;

    /**
     * The last player yaw angle. Used when free looking
     *
     * @see Settings#freeLook
     */
    private float lastYaw;

    public LookBehavior(Baritone baritone) {
        super(baritone);
    }

    @Override
    public void updateTarget(Rotation target, boolean force) {
        this.target = target;
        this.force = force || !Baritone.settings().freeLook.value;
    }

    // TODO: for some reason with freecam on this breaks
    @Override
    public void onPlayerUpdate(PlayerUpdateEvent event) {
        if (this.target == null) {
            return;
        }

        // Whether or not we're going to silently set our angles
        boolean silent = Baritone.settings().antiCheatCompatibility.value && !this.force;

        switch (event.getState()) {
            case PRE: {
                if (this.force) {
                    float desiredPitch = this.target.getPitch();
                    float desiredYaw = this.target.getYaw();

                    if (!Baritone.settings().freeLook.value) {
                        lookTo(desiredYaw, desiredPitch);
                    }

                    this.target = null;
                }

                if (silent) {
                    this.lastYaw = ctx.player().rotationYaw;
                    ctx.player().rotationPitch = this.target.getPitch();
                    ctx.player().rotationYaw = this.target.getYaw();
                }

                break;
            }
            case POST: {
                if (silent) {
                    ctx.player().rotationYaw = this.lastYaw;
                    this.target = null;
                }
                break;
            }
            default:
                break;
        }
    }

    public void pig() {
        if (this.target != null) {
            ctx.player().rotationYaw = this.target.getYaw();
        }
    }

    @Override
    public void onPlayerRotationMove(RotationMoveEvent event) {
        if (this.target != null) {

            event.setYaw(this.target.getYaw());

            // If we have antiCheatCompatibility on, we're going to use the target value later in onPlayerUpdate()
            // Also the type has to be MOTION_UPDATE because that is called after JUMP
            if (!Baritone.settings().antiCheatCompatibility.value && event.getType() == RotationMoveEvent.Type.MOTION_UPDATE && !this.force) {
                this.target = null;
            }
        }
    }

    /**
     * Smoothly pans to desired rotation
     * @param desiredYaw
     * @param desiredPitch
     */
    private void lookTo(float desiredYaw, float desiredPitch) {
        if (desiredPitch == ctx.player().rotationPitch) {
            desiredPitch = 0;
        }

        float dy = desiredYaw - ctx.player().rotationYaw;
        float dp = desiredPitch - ctx.player().rotationPitch;
        float inc = Baritone.settings().lookToIncrement.value;

        if (dy > -inc && dy < inc) {
            ctx.player().rotationYaw += dy;
        } else {
            ctx.player().rotationYaw += inc * MathHelper.clamp(dy, -1, 1);
        }

        if (dp > -inc && dp < inc) {
            ctx.player().rotationPitch += dp;
        } else {
            ctx.player().rotationPitch += inc * MathHelper.clamp(dp, -1, 1);
        }

        ctx.player().rotationYaw += (Math.random() - 0.5) * Baritone.settings().randomLooking.value;
        ctx.player().rotationPitch += (Math.random() - 0.5) * Baritone.settings().randomLooking.value;
    }
}
